import logging

from pathlib import Path
from datetime import datetime
from urllib.parse import urlencode

from scrapy.spiders import Spider
from scrapy.signals import response_received
from scrapy.http.request import Request as BaseRequest

from .responses import TextResponse, HtmlResponse, XmlResponse, JsonResponse, \
                       RISResponse


__all__ = ['Scraper', 'Request', 'TextResponse', 'HtmlResponse', 'XmlResponse',
           'JsonResponse', 'RISResponse']


class Request(BaseRequest):
    def __init__(self, url, callback=None, method='GET', headers=None, body=None,
             cookies=None, meta=None, encoding='utf-8', priority=0,
             dont_filter=False, errback=None, flags=None, params=None):

        if params:
            url += '?' + urlencode(params)

        super().__init__(url, callback, method, headers, body, cookies, meta,
                encoding, priority, dont_filter, errback, flags)


class Scraper(Spider):
    MAX_COUNT = None
    BATCH_SIZE = 100
    DEBUG_DIR = '/tmp'
    ARGUMENTS = ['url', 'count', 'query', 'count_only']

    url = None
    count = None
    query = ''
    count_only = False

    def __init__(self, *args, **kwargs):
        # The default __init__ method will take any spider arguments and copy
        # them to the spider as attributes: filter arguments for security
        # purposes.
        spider_args = {k: v for k, v in kwargs.items() if k in self.ARGUMENTS}

        super().__init__(*args, **spider_args)

        default_parser = getattr(self, 'default_parser', None)
        if default_parser and not hasattr(self, 'parse'):
            # XXX Use setattr to bypass pylint warning...
            setattr(self, 'parser', getattr(self, default_parser))

    def start_requests(self):
        if self.url: # and self.url.startswith('file://'):
            yield Request(self.url)
        else:
            yield from self.dispatch()

    @property
    def logger_name(self):
        return 'scrapers.%s' % self.name

    @property
    def logger(self):
        logger = logging.getLogger(self.logger_name)
        return logging.LoggerAdapter(logger, {'spider': self})

    @property
    def limit(self):
        if self.MAX_COUNT is None:
            return self.count or 0
        if self.count is None:
            return self.MAX_COUNT
        return min(self.count, self.MAX_COUNT)

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = super().from_crawler(crawler, *args, **kwargs)
        crawler.signals.connect(spider.trace, signal=response_received)
        return spider

    def trace(self, response, request, spider):
        content_type = response.headers.get('content-type', b'').decode()

        self.logger.info('Content-Type=%s; type(response)=%s;',
            content_type, type(response).__name__)

        path = Path(self.DEBUG_DIR).absolute()
        date = datetime.now().strftime("%Y%m%d_%H%m_%s.%f")
        ext = '.html' if isinstance(response, HtmlResponse) else \
              '.xml' if isinstance(response, XmlResponse) else \
              '.json' if isinstance(response, JsonResponse) else \
              '.txt' if isinstance(response, TextResponse) else \
              ''

        filename = '%s-%s%s' % (spider.logger_name, date, ext)
        filepath = str(path / filename)

        with open(filepath, 'wb') as f:
            f.write(response.body)
