import logging
import jmespath

from abc import ABC
from weakref import WeakValueDictionary

from scrapy.http import \
    TextResponse as BaseText, HtmlResponse as BaseHtml, XmlResponse as BaseXml
from scrapy import responsetypes as _responsetypes
from scrapy.responsetypes import ResponseTypes as BaseResponseTypes

from gargantext.utils.json import json_loads

from RISparser.parser import Ris
from RISparser.config import TAG_KEY_MAPPING


logger = logging.getLogger('scrapers')


# To be used in conjunction with gargantext.datasource.file.FileDownloadHandler
class StreamableMixin(object):
    def __init__(self, *args, **kwargs):
        self.stream = kwargs.pop('stream', None)
        self._cached_stream_data = None

        super().__init__(*args, **kwargs)

    def readlines(self):
        return iter(self.stream)

    def _get_body(self):
        if self.stream is not None:
            if self._cached_stream_data is None:
                self._cached_stream_data = self.stream.read()
            return self._cached_stream_data

        return super()._get_body()


class TextResponse(StreamableMixin, BaseText, ABC):
    pass


class HtmlResponse(StreamableMixin, BaseHtml): pass
TextResponse.register(HtmlResponse)


class XmlResponse(StreamableMixin, BaseXml): pass
TextResponse.register(XmlResponse)


class ParseableResponse(TextResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_data = None
        self._jmes_cache = WeakValueDictionary()

    def parse(self):
        raise NotImplementedError("don't use ParseableResponse directly")

    @property
    def data(self):
        if self._cached_data is None:
            self._cached_data = self.parse()
        return self._cached_data

    def jmes(self, path):
        jp = self._jmes_cache.get(path)

        if jp is None:
            jp = self._jmes_cache[path] = jmespath.compile(path)

        return jp.search(self.data)


class JsonResponse(ParseableResponse):
    def parse(self):
        return json_loads(self.text)


class RISResponse(ParseableResponse):
    class RIS(Ris):
        PATTERN = '^[A-Z][A-Z0-9]  -'

        def __init__(self, lines):
            super().__init__(lines, TAG_KEY_MAPPING)

    def parse(self):
        return self.RIS(self.readlines()).parse()


class ResponseTypes(BaseResponseTypes):
    CLASSES = {
        'text/html': 'gargantext.datasource.responses.HtmlResponse',
        'application/atom+xml': 'gargantext.datasource.responses.XmlResponse',
        'application/rdf+xml': 'gargantext.datasource.responses.XmlResponse',
        'application/rss+xml': 'gargantext.datasource.responses.XmlResponse',
        'application/xhtml+xml': 'gargantext.datasource.responses.HtmlResponse',
        'application/vnd.wap.xhtml+xml': 'gargantext.datasource.responses.HtmlResponse',
        'application/xml': 'gargantext.datasource.responses.XmlResponse',
        'application/json': 'gargantext.datasource.responses.JsonResponse',
        'application/x-json': 'gargantext.datasource.responses.JsonResponse',
        'application/openapi+json': 'gargantext.datasource.responses.JsonResponse',
        'application/json-amazonui-streaming': 'gargantext.datasource.responses.TextResponse',
        'application/javascript': 'gargantext.datasource.responses.TextResponse',
        'application/x-javascript': 'gargantext.datasource.responses.TextResponse',
        'text/xml': 'gargantext.datasource.responses.XmlResponse',
        'text/*': 'gargantext.datasource.responses.TextResponse',
    }

_responsetypes.responsetypes = responsetypes = ResponseTypes()
